package org.nucleus8583.core;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.io.Writer;
import java.util.Arrays;
import java.util.BitSet;

import org.nucleus8583.core.charset.spi.CharsetProvider;

/**
 * This class represents an ISO-8583 message. You can read, manipulate, and
 * write ISO-8583 message using this class.
 * 
 * To read ISO-8583 message, you can simply use: <code>
 * InputStream config = ...
 * 
 * Iso8583MessageFactory factory = Iso8583MessageFactory.create(config);
 * Iso8583Message msg = factory.createMessage();
 * 
 * InputStream raw = ...
 * msg.unpack(raw);
 * </code>
 * 
 * To write ISO-8583 message, you can simply use: <code>
 * Iso8583Message msg = ...
 * 
 * OutputStream raw = ...
 * msg.pack(raw);
 * </code>
 * 
 * @author Robbi Kurniawan
 * 
 */
public final class Iso8583Message {
	private final Iso8583Field[] fields;

	private final int count;

	private final boolean[] binaries;

	private final CharsetProvider charsetProvider;

	private String mti;

	private final String[] stringValues;

	private final BitSet[] binaryValues;

	private final BitSet bits1To128;

	private final BitSet bits129To192;

	public Iso8583Message(Iso8583Field[] fields, int count, boolean[] binaries,
			CharsetProvider charsetProvider) {
		this.fields = fields;
		this.count = count;

		this.binaries = binaries;
		this.stringValues = new String[count];
		this.binaryValues = new BitSet[count];

		this.bits1To128 = new BitSet(128);
		this.bits129To192 = new BitSet(64);

		this.charsetProvider = charsetProvider;
	}

	/**
	 * set MTI field value, field number 0 in standard ISO-8583 message
	 * 
	 * @param value
	 *            new MTI field value
	 */
	public void setMti(String mti) {
		this.mti = mti;
	}

	/**
	 * clear MTI field value, field number 0 in standard ISO-8583 message
	 */
	public void unsetMti() {
		mti = null;
	}

	/**
	 * retrieve MTI field value, field number 0 in standard ISO-8583 message
	 * 
	 * @return MTI field value
	 */
	public String getMti() {
		return mti;
	}

	/**
	 * set binary (<code>b</code> data element) field value.
	 * 
	 * @param no
	 *            number of field to be set, should in range <code>2-192</code>,
	 *            exclude <code>65<code>.
	 * @param value
	 *            new binary value
	 * @return <code>false<code> if <code>no</code> less than <code>2</code> or
	 *         more than <code>192</code> or equals to <code>65</code> or more
	 *         than number of fields defined in configuration used by
	 *         {@link Iso8583MessageFactory} that instantiates this object.
	 *         otherwise <code>true</code>
	 */
	public boolean set(int no, BitSet value) {
		if ((no <= 1) || (no > 192) || (no == 65) || (no > fields.length)) {
			return false;
		}

		if (!binaries[no]) {
			return false;
		}

		binaryValues[no] = value;
		stringValues[no] = null;

		if (no > 128) {
			bits129To192.set(no - 129);
		} else {
			bits1To128.set(no - 1);
		}

		return true;
	}

	/**
	 * set string-based (non <code>b</code> data element) field value.
	 * 
	 * @param no
	 *            number of field to be set, should in range <code>2-192</code>,
	 *            exclude <code>65<code>.
	 * @param value
	 *            new value
	 * @return <code>false<code> if <code>no</code> less than <code>2</code> or
	 *         more than <code>192</code> or equals to <code>65</code> or more
	 *         than number of fields defined in configuration used by
	 *         {@link Iso8583MessageFactory} that instantiates this object.
	 *         otherwise <code>true</code>
	 */
	public boolean set(int no, String value) {
		if ((no <= 1) || (no > 192) || (no == 65) || (no > fields.length)) {
			return false;
		}

		if (binaries[no]) {
			return false;
		}

		binaryValues[no] = null;
		stringValues[no] = value;

		if (no > 128) {
			bits129To192.set(no - 129);
		} else {
			bits1To128.set(no - 1);
		}

		return true;
	}

	/**
	 * set binary (<code>b</code> data element) field value.
	 * 
	 * this is unsafe method since no range checking performed so <b>PLEASE USE
	 * THIS METHOD WITH CARE</b>.
	 * 
	 * @param no
	 *            number of field to be set, should in range <code>2-192</code>,
	 *            exclude <code>65<code>.
	 * @param value
	 *            new binary value
	 */
	public void unsafeSet(int no, String value) {
		binaryValues[no] = null;
		stringValues[no] = value;

		if (no > 128) {
			bits129To192.set(no - 129);
		} else {
			bits1To128.set(no - 1);
		}
	}

	/**
	 * set string-based (non <code>b</code> data element) field value.
	 * 
	 * this is unsafe method since no range checking performed so <b>PLEASE USE
	 * THIS METHOD WITH CARE</b>.
	 * 
	 * @param no
	 *            number of field to be set, should in range <code>2-192</code>,
	 *            exclude <code>65<code>.
	 * @param value
	 *            new value
	 */
	public void unsafeSet(int no, BitSet value) {
		binaryValues[no] = value;
		stringValues[no] = null;

		if (no > 128) {
			bits129To192.set(no - 129);
		} else {
			bits1To128.set(no - 1);
		}
	}

	/**
	 * clear field value
	 * 
	 * @param no
	 *            number of field to be cleared, should in range
	 *            <code>2-192</code>, exclude <code>65<code>.
	 * @return <code>false<code> if <code>no</code> less than <code>2</code> or
	 *         more than <code>192</code> or equals to <code>65</code> or more
	 *         than number of fields defined in configuration used by
	 *         {@link Iso8583MessageFactory} that instantiates this object.
	 *         otherwise <code>true</code>
	 */
	public boolean unset(int no) {
		if ((no <= 1) || (no > 192) || (no == 65) || (no > fields.length)) {
			return false;
		}

		if (no > 128) {
			bits129To192.clear(no - 129);
		} else {
			bits1To128.clear(no - 1);
		}

		// let gc do it's work
		binaryValues[no] = null;
		stringValues[no] = null;

		return true;
	}

	/**
	 * clear field value.
	 * 
	 * this is unsafe method since no range checking performed so <b>PLEASE USE
	 * THIS METHOD WITH CARE</b>.
	 * 
	 * @param no
	 *            number of field to be cleared, should in range
	 *            <code>2-192</code>, exclude <code>65<code>.
	 * @param value
	 *            new value
	 */
	public void unsafeUnset(int no) {
		if (no > 128) {
			bits129To192.clear(no - 129);
		} else {
			bits1To128.clear(no - 1);
		}

		// let gc do it's work
		binaryValues[no] = null;
		stringValues[no] = null;
	}

	/**
	 * retrieve field value
	 * 
	 * @param no
	 *            number of field to be get, should in range <code>2-192</code>,
	 *            exclude <code>65<code>.
	 * @return <code>null<code> if <code>no</code> less than <code>2</code> or
	 *         more than <code>192</code> or equals to <code>65</code> or more
	 *         than number of fields defined in configuration used by
	 *         {@link Iso8583MessageFactory} that instantiates this object.
	 *         otherwise the field value.
	 */
	public Object get(int no) {
		if ((no <= 1) || (no > 192) || (no == 65) || (no > fields.length)) {
			return null;
		}

		if (binaries[no]) {
			return binaryValues[no];
		}

		return stringValues[no];
	}

	/**
	 * retrieve non <code>b</code> data element field value.
	 * 
	 * @param no
	 *            number of field to be get, should in range <code>2-192</code>,
	 *            exclude <code>65<code>.
	 * @return <code>null<code> if <code>no</code> less than <code>2</code> or
	 *         more than <code>192</code> or equals to <code>65</code> or more
	 *         than number of fields defined in configuration used by
	 *         {@link Iso8583MessageFactory} that instantiates this object or
	 *         the field has binary (<code>b</code> data element) type.
	 *         otherwise the field value.
	 */
	public String getString(int no) {
		if ((no <= 1) || (no > 192) || (no == 65) || (no > fields.length)) {
			return null;
		}

		return stringValues[no];
	}

	/**
	 * retrieve <code>b</code> data element field value.
	 * 
	 * @param no
	 *            number of field to be get, should in range <code>2-192</code>,
	 *            exclude <code>65<code>.
	 * @return <code>null<code> if <code>no</code> less than <code>2</code> or
	 *         more than <code>192</code> or equals to <code>65</code> or more
	 *         than number of fields defined in configuration used by
	 *         {@link Iso8583MessageFactory} that instantiates this object or
	 *         the field type is not binary (non <code>b</code> data element).
	 *         otherwise the field value.
	 */
	public BitSet getBinary(int no) {
		if ((no <= 1) || (no > 192) || (no == 65) || (no > fields.length)) {
			return null;
		}

		return binaryValues[no];
	}

	/**
	 * retrieve non <code>b</code> data element field value.
	 * 
	 * this is unsafe method since no range checking performed so <b>PLEASE USE
	 * THIS METHOD WITH CARE</b>.
	 * 
	 * @param no
	 *            number of field to be get, should in range <code>2-192</code>,
	 *            exclude <code>65<code>.
	 * @return <code>null<code> if the field has binary (<code>b</code> data
	 *         element) type. otherwise the field value.
	 */
	public String unsafeGetString(int no) {
		return stringValues[no];
	}

	/**
	 * retrieve <code>b</code> data element field value.
	 * 
	 * this is unsafe method since no range checking performed so <b>PLEASE USE
	 * THIS METHOD WITH CARE</b>.
	 * 
	 * @param no
	 *            number of field to be get, should in range <code>2-192</code>,
	 *            exclude <code>65<code>.
	 * @return <code>null<code> if the field type is not binary (non <code>b
	 *         </code> data element). otherwise the field value.
	 */
	public BitSet unsafeGetBinary(int no) {
		return binaryValues[no];
	}

	/**
	 * clear all fields value
	 */
	public void clear() {
		mti = null;

		Arrays.fill(binaryValues, null);
		Arrays.fill(stringValues, null);

		bits1To128.clear();
		bits129To192.clear();
	}

	/**
	 * convert this object into standard ISO-8583 message and return a byte
	 * array containing the message. This method is slower than
	 * <code>void pack(OutputStream)</code> method.
	 * 
	 * @return the byte array containing the ISO-8583 message
	 */
	public byte[] pack() {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		try {
			pack(baos);
		} catch (IOException e) {
			// should not be here
		}

		return baos.toByteArray();
	}

	/**
	 * convert this object into ISO-8583 message and write to given
	 * {@link OutputStream}. This method is slower than
	 * <code>void pack(Writer)</code> method.
	 * 
	 * @param out
	 *            the {@link OutputStream}
	 * @throws IOException
	 *             thrown if an IO error occured while converting.
	 */
	public void pack(OutputStream out) throws IOException {
		pack(charsetProvider.createEncoder(out));
	}

	/**
	 * convert this object into ISO-8583 message and write to given
	 * {@link Writer}.
	 * 
	 * @param writer
	 *            the {@link Writer}
	 * @throws IOException
	 *             thrown if an IO error occured while converting.
	 */
	public void pack(Writer writer) throws IOException {
		if (bits129To192.isEmpty()) {
			bits1To128.clear(64);

			binaryValues[65] = null;
			stringValues[65] = null;
		} else {
			bits1To128.set(64);

			binaryValues[65] = bits129To192;
			stringValues[65] = null;
		}

		fields[0].pack(writer, mti);
		fields[1].pack(writer, bits1To128);

		for (int i = 2, j = 1; (i < count) && (i < 129); ++i) {
			if (bits1To128.get(j++)) {
				if (binaries[i]) {
					fields[i].pack(writer, binaryValues[i]);
				} else {
					fields[i].pack(writer, stringValues[i]);
				}
			}
		}

		for (int i = 129, j = 0; i < count; ++i) {
			if (bits129To192.get(j++)) {
				if (binaries[i]) {
					fields[i].pack(writer, binaryValues[i]);
				} else {
					fields[i].pack(writer, stringValues[i]);
				}
			}
		}
	}

	/**
	 * read ISO-8583 message contained in given byte array and assign the fields
	 * value to this object. This method is slower than
	 * <code>void unpack(InputStream)</code> method.
	 * 
	 * @param in
	 *            the byte array
	 */
	public void unpack(byte[] in) {
		try {
			unpack(new ByteArrayInputStream(in));
		} catch (IOException ex) {
			// should not be here
		}
	}

	/**
	 * read ISO-8583 message from given {@link InputStream} and assign the
	 * fields value to this object. This method is slower than
	 * <code>void unpack(Reader)</code> method.
	 * 
	 * @param in
	 *            the {@link InputStream}
	 * @throws IOException
	 *             thrown if an IO error occured while converting.
	 */
	public void unpack(InputStream in) throws IOException {
		unpack(charsetProvider.createDecoder(in));
	}

	/**
	 * read ISO-8583 message from given {@link Reader} and assign the fields
	 * value to this object.
	 * 
	 * @param in
	 *            the {@link Reader}
	 * @throws IOException
	 *             thrown if an IO error occured while converting.
	 */
	public void unpack(Reader in) throws IOException {
		mti = fields[0].unpackString(in);
		fields[1].unpackBinary(in, bits1To128);

		// read bit-by-bit
		for (int i = 2, iMin1 = 1, iMin129 = -127; i < count; ++i, ++iMin1, ++iMin129) {
			if (i == 65) {
				if (bits1To128.get(64)) {
					fields[i].unpackBinary(in, bits129To192);
				}
			} else if (i < 129) {
				if (bits1To128.get(iMin1)) {
					if (binaries[i]) {
						unsafeSet(i, fields[i].unpackBinary(in));
					} else {
						unsafeSet(i, fields[i].unpackString(in));
					}
				}
			} else {
				if (bits129To192.get(iMin129)) {
					if (binaries[i]) {
						unsafeSet(i, fields[i].unpackBinary(in));
					} else {
						unsafeSet(i, fields[i].unpackString(in));
					}
				}
			}
		}
	}

	private boolean equals(Object a, Object b) {
		if (a == b) {
			return true;
		}

		if ((a == null) || (b == null)) {
			return false;
		}

		return a.equals(b);
	}

	@Override
	public boolean equals(Object object) {
		if (object == null) {
			return false;
		}

		if (!(object instanceof Iso8583Message)) {
			return false;
		}

		Iso8583Message another = (Iso8583Message) object;
		if (!equals(this.mti, another.mti)) {
			return false;
		}

		if (this.count != another.count) {
			return false;
		}

		for (int i = this.count - 1; i >= 2; --i) {
			if (i == 65) {
				continue;
			}

			if (this.binaries[i] != another.binaries[i]) {
				return false;
			}

			if (this.binaries[i]) {
				if (!equals(this.binaryValues[i], another.binaryValues[i])) {
					return false;
				}
			} else {
				if (!equals(this.stringValues[i], another.stringValues[i])) {
					return false;
				}
			}
		}

		return true;
	}

	@Override
	public int hashCode() {
		return toString().hashCode();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		StringBuffer sbuf = new StringBuffer();

		sbuf.append("<iso-message>\n");
		sbuf.append("    <iso-field id=\"0\" value=");
		if (mti == null) {
			sbuf.append("null");
		} else {
			sbuf.append('"');
			sbuf.append(mti);
			sbuf.append('"');
		}
		sbuf.append(" />\n");

		for (int i = 2, j = 1; (i < fields.length) && (i < 129); ++i, ++j) {
			if (bits1To128.get(j)) {
				sbuf.append("    <iso-field id=\"");
				sbuf.append(i);
				sbuf.append("\" value=");

				if (binaries[i]) {
					BitSet binaryValue = binaryValues[i];
					if (binaryValue == null) {
						sbuf.append("null");
					} else {
						sbuf.append('"');
						sbuf.append(binaryValue);
						sbuf.append('"');
					}
				} else {
					String stringValue = stringValues[i];
					if (stringValue == null) {
						sbuf.append("null");
					} else {
						sbuf.append('"');
						sbuf.append(stringValue);
						sbuf.append('"');
					}
				}

				sbuf.append(" />\n");
			}
		}

		for (int i = 129, j = 0; i < fields.length; ++i, ++j) {
			if (bits129To192.get(j)) {
				sbuf.append("    <iso-field id=\"");
				sbuf.append(i);
				sbuf.append("\" value=");

				if (binaries[i]) {
					BitSet binaryValue = binaryValues[i];
					if (binaryValue == null) {
						sbuf.append("null");
					} else {
						sbuf.append('"');
						sbuf.append(binaryValue);
						sbuf.append('"');
					}
				} else {
					String stringValue = stringValues[i];
					if (stringValue == null) {
						sbuf.append("null");
					} else {
						sbuf.append('"');
						sbuf.append(stringValue);
						sbuf.append('"');
					}
				}

				sbuf.append(" />\n");
			}
		}
		sbuf.append("</iso-message>\n");

		return sbuf.toString();
	}
}
